# calculate dicodon/5mer
import os
import numpy as np
np_load_old = np.load
np.load = lambda *a,**k: np_load_old(*a, allow_pickle=True, **k)
import pandas as pd
from Bio import SeqIO
from kpal.klib import Profile
import time
from numba import jit

os.chdir("/data/jtan/HoPhage_revise/data")

d = dict({'A':0,'X':0,'D':0,'W':0,
          'C':1,'M':1,'S':1,'B':1,
          'G':2,'N':2,'K':2,'R':2,'V':2,
          'T':3,'H':3,'Y':3})

def str_to_num(seq,d):
    seq_num = np.zeros(len(seq),dtype=int)
    for i in range(len(seq)):
        seq_num[i] = d[seq[i]]
    return seq_num

@jit(nopython=True)
def calc(seq_num,Dicodon):
    lin = np.arange(0, int(len(seq_num) / 3) * 3, 3)
    codon = seq_num[lin] * 16 + \
            seq_num[lin + 1] * 4 + seq_num[lin + 2]
    a = codon[:-1]
    b = codon[1:]
    for n in range(len(a)):
        Dicodon[a[n], b[n]] = Dicodon[a[n], b[n]] + 1

    return Dicodon

@jit(nopython=True)
def cals(seq_num,mer):
    temp = seq_num[:-4] * 256 + seq_num[1:-3] * 64 + seq_num[2:-2] * 16 \
           + seq_num[3:-1] * 4 + seq_num[4:]
    for i in range(len(temp)):
        mer[temp[i]] = mer[temp[i]] + 1
    return mer

gb_folder_a = 'refseq_archaea/'
gb_folder_b = 'refseq_bacteria/'

host_info_a = pd.read_csv("info_reseq_archaea.csv",sep='\t',keep_default_na=False)
host_info_b = pd.read_csv("info_reseq_bacteria.csv",sep='\t',keep_default_na=False)
genera_a = np.unique(host_info_a.genus)
genera_b = np.unique(host_info_b.genus)
MER = []
DICODON = []

for i in range(0,len(genera_a)):
    t1 = time.clock()
    print(i)
    info = host_info_a[host_info_a.genus==genera_a[i]].reset_index(drop=True)
    Mer = np.zeros(1024,dtype=int)
    Dicodon = np.zeros((64,64),dtype=int)
    for j in range(len(info)):
        # record = SeqIO.read(gb_folder_a+info.file[j], "genbank")  
        records = list(SeqIO.parse(gb_folder_a+info.file[j], "genbank"))
        for n in range(len(records)):
            record = records[n]
            seq = record.seq
            seq_num = str_to_num(seq, d)
            seq_rev_num = 3 - seq_num

            mer = np.zeros(1024, dtype=int)
            mer = cals(seq_num,mer)
            Mer = Mer + mer

            for feature in record.features:
                if feature.type == 'CDS':
                    loc = feature.location
                    if len(loc.parts) == 1:
                        # strand=-1 means complementary strand
                        if loc.strand == 1:
                            seq_numc = seq_num[loc.nofuzzy_start:loc.nofuzzy_end]
                        else:
                            seq_numc = seq_rev_num[loc.nofuzzy_start:loc.nofuzzy_end][::-1]
                    else:
                        flag = 1
                        parts = loc.parts
                        if loc.strand == 1:
                            seq_numc = np.hstack((seq_num[parts[0].nofuzzy_start:parts[0].nofuzzy_end],
                                                  seq_num[parts[1].nofuzzy_start:parts[1].nofuzzy_end]))
                        else:
                            seq_numc = np.hstack((seq_rev_num[parts[0].nofuzzy_start:parts[0].nofuzzy_end][::-1],
                                                  seq_rev_num[parts[1].nofuzzy_start:parts[1].nofuzzy_end][::-1]))
                    temp = len(seq_numc) % 3
                    if temp != 0:
                        # print(len(seq_numc))
                        if (seq_numc[-3:] == np.array([3, 0, 0])).all() or (seq_numc[-3:] == np.array([3, 0, 2])).all() or (
                                seq_numc[-3:] == np.array([3, 2, 0])).all():
                            seq_numc = seq_numc[temp:]
                            # print(len(seq_numc))

                        temp = len(seq_numc) % 3
                        if temp != 0:
                            if (seq_numc[:3] == np.array([0, 3, 2])).all() or (
                                    seq_numc[:3] == np.array([2, 3, 2])).all() or (
                                    seq_numc[:3] == np.array([3, 3, 2])).all():
                                seq_numc = seq_numc[:-temp]
                                # print(len(seq_numc))
                    temp = len(seq_numc) % 3
                    if temp == 0:
                        Dicodon = calc(seq_numc,Dicodon)
    MER.append(Mer)
    DICODON.append(Dicodon)
    t2 = time.clock()
    print("time:",(t2-t1))


for i in range(1,len(genera_b)):
    t1 = time.clock()
    print(i)
    info = host_info_b[host_info_b.genus==genera_b[i]].reset_index(drop=True)
    Mer = np.zeros(1024,dtype=int)
    Dicodon = np.zeros((64,64),dtype=int)
    for j in range(len(info)):
        # record = SeqIO.read(gb_folder_b+info.file[j], "genbank")  
        records = list(SeqIO.parse(gb_folder_b+info.file[j], "genbank"))
        for n in range(len(records)):
            record = records[n]
            seq = record.seq
            seq_num = str_to_num(seq, d)
            seq_rev_num = 3 - seq_num

            mer = np.zeros(1024, dtype=int)
            mer = cals(seq_num,mer)
            Mer = Mer + mer

            for feature in record.features:
                if feature.type == 'CDS':
                    loc = feature.location
                    if len(loc.parts) == 1:
                        # strand=-1 means complementary strand
                        if loc.strand == 1:
                            seq_numc = seq_num[loc.nofuzzy_start:loc.nofuzzy_end]
                        else:
                            seq_numc = seq_rev_num[loc.nofuzzy_start:loc.nofuzzy_end][::-1]
                    else:
                        flag = 1
                        parts = loc.parts
                        if loc.strand == 1:
                            seq_numc = np.hstack((seq_num[parts[0].nofuzzy_start:parts[0].nofuzzy_end],
                                                  seq_num[parts[1].nofuzzy_start:parts[1].nofuzzy_end]))
                        else:
                            seq_numc = np.hstack((seq_rev_num[parts[0].nofuzzy_start:parts[0].nofuzzy_end][::-1],
                                                  seq_rev_num[parts[1].nofuzzy_start:parts[1].nofuzzy_end][::-1]))
                    temp = len(seq_numc) % 3
                    if temp != 0:
                        # print(len(seq_numc))
                        if (seq_numc[-3:] == np.array([3, 0, 0])).all() or (seq_numc[-3:] == np.array([3, 0, 2])).all() or (
                                seq_numc[-3:] == np.array([3, 2, 0])).all():
                            seq_numc = seq_numc[temp:]
                            # print(len(seq_numc))

                        temp = len(seq_numc) % 3
                        if temp != 0:
                            if (seq_numc[:3] == np.array([0, 3, 2])).all() or (
                                    seq_numc[:3] == np.array([2, 3, 2])).all() or (
                                    seq_numc[:3] == np.array([3, 3, 2])).all():
                                seq_numc = seq_numc[:-temp]
                                # print(len(seq_numc))
                    temp = len(seq_numc) % 3
                    if temp == 0:
                        Dicodon = calc(seq_numc,Dicodon)
    MER.append(Mer)
    DICODON.append(Dicodon)
    t2 = time.clock()
    print("time:",(t2-t1))
np.save('DICODON_origin.npy',DICODON)
np.save('MER_origin.npy',MER)

MER = np.array(MER)
DICODON = np.array(DICODON)
MER = MER/MER.sum(axis=1)[:,None]
DICODON2 = []
for i in range(len(DICODON)):
    temp = DICODON[i] + 1
    DICODON2.append(temp/temp.sum(axis=1)[:,None])
del DICODON
DICODON = np.array(DICODON2)
del DICODON2
np.save("DICODON.npy",DICODON)
np.save("MER.npy",MER)