import os
import numpy as np
np_load_old = np.load
np.load = lambda *a,**k: np_load_old(*a, allow_pickle=True, **k)
import pandas as pd
import re
import fnmatch
from Bio import SeqIO
import time
import random
from sklearn.preprocessing import OneHotEncoder
enc = OneHotEncoder(dtype='uint8')

os.chdir("/data/jtan/HoPhage_revise/data")

##################################################################################################
d = dict({'A':0,'X':0,'D':0,'W':0,
          'C':1,'M':1,'S':1,'B':1,
          'G':2,'N':2,'K':2,'R':2,'V':2,
          'T':3,'H':3,'Y':3})

def str_to_num(seq,d):
    seq_num = np.zeros(len(seq),dtype='uint8')
    for i in range(len(seq)):
        seq_num[i] = d[seq[i]]
    return seq_num

def preprocessing():
    # input
    train_file = os.listdir('train_test/train')
    prophage_ref27 = fnmatch.filter(os.listdir('prophage/prophage_ref27'), '*.fna')
    prophage_ref42 = fnmatch.filter(os.listdir('prophage/prophage_ref42'), '*.fna')
    train_file.extend(prophage_ref27)
    train_file.extend(prophage_ref42)

    records = list(SeqIO.parse("phage_vhdb.gb", "genbank"))
    Accession = []
    for i in range(len(records)):
        Accession.append(records[i].id.split('.')[0])
    Accession = np.array(Accession)

    # output
    Train_GI = []
    Train_seq_all = [[] for i in range(len(train_file))]
    Train_seq_01 = [[] for i in range(len(train_file))]

    for i in range(len(train_file)):
        print(i)
        if not re.match('prophage',train_file[i]):
            acession = train_file[i].split('-')[0]
            ind = np.where(Accession == acession)[0]
            record = records[ind[0]]
            gi = record.annotations['gi']
            Train_GI.append(int(gi))

            seq = record.seq
            seq_num = str_to_num(seq, d)
            seq_com_num = 3 - seq_num
            Train_seq_all[i] = np.array([seq_num, seq_com_num])

            seq_01 = np.zeros((2, len(seq_num)), dtype='uint8')
            cds_num = 0
            for feature in record.features:
                if feature.type == 'CDS':
                    cds_num += 1
                    loc = feature.location
                    for part in loc.parts:
                        # strand=-1 means complementary strand
                        if part.strand == 1:
                            seq_01[0, part.start:part.end] = 1
                        else:
                            seq_01[1, part.start:part.end] = 1

            if cds_num == 0:
                # print(acession)
                temppath = "/data/jtan/mydata/phage/phagewithoutcds/"
                if not os.path.exists(temppath + acession + ".txt"):
                    SeqIO.write(record, temppath + acession + ".fna", "fasta")
                    os.system(
                        '/data2/jtan/PHANOTATE/predict.sh ' + temppath + acession + '.fna ' + temppath + acession + '.txt')
                with open(temppath + acession + '.txt') as f:
                    cds_location = f.readlines()
                for j in range(2, len(cds_location)):
                    temp = cds_location[j].split('\t')
                    # strand=-1 means complementary strand
                    s = int(temp[0])
                    e = int(temp[1])
                    if temp[2] == '+':
                        seq_01[0, s - 1:e] = 1
                    else:
                        seq_01[1, e - 1:s] = 1

            Train_seq_01[i] = seq_01
        else:
            # prophage
            if train_file[i] in prophage_ref27:
                record = SeqIO.read('prophage/prophage_ref27/'+train_file[i],'fasta')
                temp = 'prophage/prophage_ref27/seq_01_' + train_file[i][:-4] + '.npy'
            else:
                record = SeqIO.read('prophage/prophage_ref42/' + train_file[i],'fasta')
                temp = 'prophage/prophage_ref42/seq_01_' + train_file[i][:-4] + '.npy'

            seq = record.seq
            seq_num = str_to_num(seq, d)
            seq_com_num = 3 - seq_num
            Train_seq_all[i] = np.array([seq_num, seq_com_num])
            seq_01 = np.load(temp)
            Train_seq_01[i] = seq_01
            gi = train_file[i][:-4]
            Train_GI.append(gi)

    return Train_GI,Train_seq_all,Train_seq_01
Train_phage_GI,Train_phage_all,Train_phage_01 = preprocessing()
Train_phage_GI = np.array(Train_phage_GI)

info_vhdb = pd.read_csv('info_phage_host_interaction.csv')
info_vhdb = info_vhdb.drop_duplicates(subset=['phage_taxid', 'host_genus']).reset_index(drop=True)
info_ref27 = pd.read_csv('prophage/info_prophage_ref27_host.csv')
info_ref42 = pd.read_csv('prophage/info_prophage_ref42_host.csv')

###########
# host_info_a = pd.read_csv("info_reseq_archaea.csv",delimiter='\t',keep_default_na=False)
# host_info_b = pd.read_csv("info_reseq_bacteria.csv",delimiter='\t',keep_default_na=False)
# host_info = pd.concat([host_info_a,host_info_b])
# host_info = host_info.drop_duplicates(subset=['genus'])
# host_info = host_info.sort_values(by='genus',axis=0)
# host_taxa = host_info.iloc[1:,3:]
# host_taxa = host_taxa.reset_index(drop=True)
# host_taxa.to_csv("HoPhage-G/host_taxa_1353.csv",index=False)
host_taxa = pd.read_csv("HoPhage-G/host_taxa_1353.csv")
###########

# fasta_train = list(SeqIO.parse("train_test/metasim_train/train-100-400.fna", "fasta"))
# fasta_train = list(SeqIO.parse("train_test/metasim_train/train-400-800.fna", "fasta"))
fasta_train = list(SeqIO.parse("train_test/metasim_train/train-800-1200.fna", "fasta"))

Input_train_seq = []
Input_train_01 = []
Input_train_label = []
# for i in range(1):
for i in range(len(fasta_train)):
    if i%10000==0:
        print(i)
    desc = fasta_train[i].description

    if 'GI' in desc:
        info = re.findall('{GI=.*?}', desc)[0]
        temp = info[1:-1].split(',')
        gi = temp[0][3:]
    else:
        info = re.findall('{KEY=.*?}', desc)[0]
        temp = info[1:-1].split(',')

        temp2 = re.findall('prophage.*? ', desc)[0]
        gi = temp2[:-1] 

    fw_bw = temp[1]
    position = temp[2].split('-')
    ind = position[1].find(';')
    if ind!=-1:
        position2 = temp[4].split('-')
        if fw_bw=='fw':
            s = [int(position[0]),int(position2[0])]
            e = [int(position[1][:ind]),int(position2[1])]
        else:
            s = [int(position2[0]), int(position[0])]
            e = [int(position2[1]), int(position[1][:ind])]
    else:
        s = [int(position[0])]
        e = [int(position[1])]

    
    ind = np.where(Train_phage_GI==gi)[0][0]
    phage_seq = Train_phage_all[ind]
    seq_num = [phage_seq[0, s[0]:e[0]], phage_seq[1, s[0]:e[0]][::-1]]
    if len(s) == 2:
        seq_num = [np.append(seq_num[0], phage_seq[0, s[1]:e[1]]),
                  np.append(seq_num[1][::-1], phage_seq[1, s[1]:e[1]])[::-1]]
    ########
    # seq_num_2 = np.ones((2,400),dtype='uint8')*4
    # seq_num_2 = np.ones((2, 800), dtype='uint8') * 4
    seq_num_2 = np.ones((2, 1200), dtype='uint8') * 4
    ########
    seq_num_2[0,:len(seq_num[0])] = seq_num[0]
    seq_num_2[1,:len(seq_num[0])] = seq_num[1]
    enc.fit([[0], [1], [2], [3], [4]])
    onehot5 = enc.transform(np.hstack((seq_num_2[0],seq_num_2[1]))[:, np.newaxis]).toarray()
    seq_onehot = onehot5[:, 0:4]

    phage_01 = Train_phage_01[ind]
    seq_01 = [phage_01[0,s[0]:e[0]],phage_01[1,s[0]:e[0]][::-1]]
    if len(s)==2:
        seq_01 = [np.append(seq_01[0],phage_01[0, s[1]:e[1]]),
                  np.append(seq_01[1][::-1],phage_01[1, s[1]:e[1]])[::-1]]
    ########
    # seq_01_2 = np.ones((2, 400), dtype='uint8') * 2
    # seq_01_2 = np.ones((2, 800), dtype='uint8') * 2
    seq_01_2 = np.ones((2, 1200), dtype='uint8') * 2
    ########
    seq_01_2[0, :len(seq_01[0])] = seq_01[0]
    seq_01_2[1, :len(seq_01[0])] = seq_01[1]
    enc.fit([[0], [1], [2]])
    onehot3 = enc.transform(np.hstack((seq_01_2[0], seq_01_2[1]))[:, np.newaxis]).toarray()
    seq01_onehot = onehot3[:, 0:2]

    # construct pairs
    if str.isalnum(gi):
        info_gi = info_vhdb[info_vhdb['phage_gi']==int(gi)]
        genus = info_gi['host_genus']
        genus = genus.drop_duplicates().values
    else:
        # ref27 or 42
        temp = gi.split('_')
        if len(temp)==4:
            accession = temp[-2] + '_' + temp[-1].split('.')[0]
            info_gi = info_ref27[info_ref27['accession'].str.contains(accession)]
        else:
            taxid = temp[-1]
            info_gi = info_ref42[info_ref42['taxid'] == int(taxid)]
        genus = info_gi['genus'].values

    train_label = np.zeros(len(host_taxa), dtype='uint8')
    ind1 = host_taxa[host_taxa['genus'].isin(genus)].index
    train_label[ind1] = 1

    seq_onehot = np.array(seq_onehot, dtype='uint8')
    seq01_onehot = np.array(seq01_onehot, dtype='uint8')
    Input_train_seq.append(seq_onehot)
    Input_train_01.append(seq01_onehot)
    Input_train_label.append(train_label)

print(len(Input_train_label))

Input_train_label = np.array(Input_train_label,dtype='uint8')
Input_train_seq = np.array(Input_train_seq,dtype='uint8')
Input_train_01 = np.array(Input_train_01,dtype='uint8')

# np.save("HoPhage-G/Input_train_label-100-400.npy",Input_train_label)
# np.save("HoPhage-G/Input_train_seq-100-400.npy",Input_train_seq)
# np.save("HoPhage-G/Input_train_01-100-400.npy",Input_train_01)

# np.save("HoPhage-G/Input_train_label-400-800.npy",Input_train_label)
# np.save("HoPhage-G/Input_train_seq-400-800.npy",Input_train_seq)
# np.save("HoPhage-G/Input_train_01-400-800.npy",Input_train_01)

np.save("HoPhage-G/Input_train_label-800-1200.npy",Input_train_label)
np.save("HoPhage-G/Input_train_seq-800-1200.npy",Input_train_seq)
np.save("HoPhage-G/Input_train_01-800-1200.npy",Input_train_01)